@@ -67,21 +67,26 @@
      VAR_XFRDFILE = 283,
      VAR_DIFFFILE = 284,
      VAR_XFRD_RELOAD_TIMEOUT = 285,
-     VAR_ZONEFILE = 286,
-     VAR_ZONE = 287,
-     VAR_ALLOW_NOTIFY = 288,
-     VAR_REQUEST_XFR = 289,
-     VAR_NOTIFY = 290,
-     VAR_PROVIDE_XFR = 291,
-     VAR_OUTGOING_INTERFACE = 292,
-     VAR_ALLOW_AXFR_FALLBACK = 293,
-     VAR_KEY = 294,
-     VAR_ALGORITHM = 295,
-     VAR_SECRET = 296,
-     VAR_AXFR = 297,
-     VAR_UDP = 298,
-     VAR_VERBOSITY = 299,
-     VAR_HIDE_VERSION = 300
+     VAR_TCP_QUERY_COUNT = 286,
+     VAR_TCP_TIMEOUT = 287,
+     VAR_IPV4_EDNS_SIZE = 288,
+     VAR_IPV6_EDNS_SIZE = 289,
+     VAR_ZONEFILE = 290,
+     VAR_ZONE = 291,
+     VAR_ALLOW_NOTIFY = 292,
+     VAR_REQUEST_XFR = 293,
+     VAR_NOTIFY = 294,
+     VAR_PROVIDE_XFR = 295,
+     VAR_NOTIFY_RETRY = 296,
+     VAR_OUTGOING_INTERFACE = 297,
+     VAR_ALLOW_AXFR_FALLBACK = 298,
+     VAR_KEY = 299,
+     VAR_ALGORITHM = 300,
+     VAR_SECRET = 301,
+     VAR_AXFR = 302,
+     VAR_UDP = 303,
+     VAR_VERBOSITY = 304,
+     VAR_HIDE_VERSION = 305
    };
 #endif
 /* Tokens.  */
@@ -113,21 +118,26 @@
 #define VAR_XFRDFILE 283
 #define VAR_DIFFFILE 284
 #define VAR_XFRD_RELOAD_TIMEOUT 285
-#define VAR_ZONEFILE 286
-#define VAR_ZONE 287
-#define VAR_ALLOW_NOTIFY 288
-#define VAR_REQUEST_XFR 289
-#define VAR_NOTIFY 290
-#define VAR_PROVIDE_XFR 291
-#define VAR_OUTGOING_INTERFACE 292
-#define VAR_ALLOW_AXFR_FALLBACK 293
-#define VAR_KEY 294
-#define VAR_ALGORITHM 295
-#define VAR_SECRET 296
-#define VAR_AXFR 297
-#define VAR_UDP 298
-#define VAR_VERBOSITY 299
-#define VAR_HIDE_VERSION 300
+#define VAR_TCP_QUERY_COUNT 286
+#define VAR_TCP_TIMEOUT 287
+#define VAR_IPV4_EDNS_SIZE 288
+#define VAR_IPV6_EDNS_SIZE 289
+#define VAR_ZONEFILE 290
+#define VAR_ZONE 291
+#define VAR_ALLOW_NOTIFY 292
+#define VAR_REQUEST_XFR 293
+#define VAR_NOTIFY 294
+#define VAR_PROVIDE_XFR 295
+#define VAR_NOTIFY_RETRY 296
+#define VAR_OUTGOING_INTERFACE 297
+#define VAR_ALLOW_AXFR_FALLBACK 298
+#define VAR_KEY 299
+#define VAR_ALGORITHM 300
+#define VAR_SECRET 301
+#define VAR_AXFR 302
+#define VAR_UDP 303
+#define VAR_VERBOSITY 304
+#define VAR_HIDE_VERSION 305
 
 
 
@@ -139,7 +149,7 @@ typedef union YYSTYPE
 	char*	str;
 }
 /* Line 1489 of yacc.c.  */
-#line 143 "configparser.h"
+#line 153 "configparser.h"
 	YYSTYPE;
 # define yystype YYSTYPE /* obsolescent; will be withdrawn */
 # define YYSTYPE_IS_DECLARED 1
