@@ -46,12 +46,9 @@
  * Exit codes are based on named-xfer for now.  See ns_defs.h in
  * bind8.
  */
-enum nsd_xfer_exit_codes
-{
-	XFER_UPTODATE = 0,
-	XFER_SUCCESS  = 1,
-	XFER_FAIL     = 3
-};
+#define XFER_UPTODATE	0
+#define XFER_SUCCESS	1
+#define XFER_FAIL		3
 
 struct axfr_state
 {
@@ -187,12 +184,13 @@ to_alarm(int ATTR_UNUSED(sig))
 	timeout_flag = 1;
 }
 
+#ifdef TSIG
 /*
  * Read a line from IN.  If successful, the line is stripped of
  * leading and trailing whitespace and non-zero is returned.
  */
 static int
-read_line(FILE *in, char *line, size_t size)
+tsig_read_line(FILE *in, char *line, size_t size)
 {
 	if (!fgets(line, size, in)) {
 		return 0;
@@ -202,7 +200,6 @@ read_line(FILE *in, char *line, size_t size)
 	}
 }
 
-#ifdef TSIG
 static tsig_key_type *
 read_tsig_key_data(region_type *region, FILE *in,
 	int ATTR_UNUSED(default_family), tsig_algorithm_type** tsig_algo)
@@ -215,7 +212,7 @@ read_tsig_key_data(region_type *region, FILE *in,
 	uint8_t data[4000];
 
 	/* server address */
-	if (!read_line(in, line, sizeof(line))) {
+	if (!tsig_read_line(in, line, sizeof(line))) {
 		error("failed to read TSIG key server address: '%s'",
 		      strerror(errno));
 		return NULL;
@@ -223,7 +220,7 @@ read_tsig_key_data(region_type *region, FILE *in,
 	/* server address unused */
 
 	/* tsig keyname */
-	if (!read_line(in, line, sizeof(line))) {
+	if (!tsig_read_line(in, line, sizeof(line))) {
 		error("failed to read TSIG key name: '%s'", strerror(errno));
 		return NULL;
 	}
@@ -235,7 +232,7 @@ read_tsig_key_data(region_type *region, FILE *in,
 	}
 
 	/* tsig algorithm */
-	if (!read_line(in, line, sizeof(line))) {
+	if (!tsig_read_line(in, line, sizeof(line))) {
 		error("failed to read TSIG key algorithm: '%s'", strerror(errno));
 		return NULL;
 	}
@@ -247,7 +244,7 @@ read_tsig_key_data(region_type *region, FILE *in,
 	}
 
 	/* tsig secret */
-	if (!read_line(in, line, sizeof(line))) {
+	if (!tsig_read_line(in, line, sizeof(line))) {
 		error("failed to read TSIG key data: '%s'\n", strerror(errno));
 		return NULL;
 	}
@@ -786,8 +783,10 @@ main(int argc, char *argv[])
 	int default_family = DEFAULT_AI_FAMILY;
 	struct sigaction mysigaction;
 	FILE *zone_file;
+#ifdef TSIG
 	const char *tsig_key_filename = NULL;
 	tsig_key_type *tsig_key = NULL;
+#endif /* TSIG */
 	axfr_state_type state;
 
 	log_init("nsd-xfer");
@@ -847,17 +846,16 @@ main(int argc, char *argv[])
 			port = optarg;
 			break;
 		case 's': {
-			long long v;
-			char *t;
+			uint32_t v;
+			const char *t;
  			state.first_transfer = 0;
-			v = strtoll(optarg, &t, 10);
-			if (optarg[0] == '\0' || *t != '\0' || v < 0 || v > UINT_MAX
-				|| (errno == ERANGE && (v == LLONG_MIN || v == LLONG_MAX)))
+			v = strtoserial(optarg, &t);
+			if (optarg[0] == '\0' || *t != '\0')
 			{
 				error("bad serial '%s'", optarg);
 				exit(XFER_FAIL);
 			}
-			state.last_serial = (uint32_t) v;
+			state.last_serial = v;
 			break;
 		}
 		case 'T':
