@@ -17,6 +17,12 @@
 #include <sys/wait.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
+#ifdef HAVE_GRP_H
+#include <grp.h>
+#endif /* HAVE_GRP_H */
+#ifdef HAVE_SETUSERCONTEXT
+#include <login_cap.h>
+#endif /* HAVE_SETUSERCONTEXT */
 
 #include <assert.h>
 #include <ctype.h>
@@ -72,7 +78,7 @@ usage (void)
 	fprintf(stderr,
 		"  -i identity     Specify the identity when queried for id.server CHAOS TXT.\n"
 #ifdef NSID
-                "  -I nsid         Specify the NSID. This must be a hex string.\n"
+		"  -I nsid         Specify the NSID. This must be a hex string.\n"
 #endif /* NSID */
 #ifndef NDEBUG
 		"  -L level        Specify the debug level.\n"
@@ -124,6 +130,20 @@ error(const char *format, ...)
 	exit(1);
 }
 
+static int
+file_inside_chroot(const char* fname, const char* chr)
+{
+#ifdef NDEBUG
+	assert(chr);
+#endif /* NDEBUG */
+	/* filename and chroot the same? */
+	if (fname && fname[0] && chr[0] && !strncmp(fname, chr, strlen(chr)))
+		return 2; /* strip chroot, file rotation ok */
+	else if (fname && fname[0] != '/')
+		return 1; /* don't strip, file rotation ok */
+	return 0; /* don't strip, don't try file rotation */
+}
+
 
 /*
  * Fetch the nsd parent process id from the nsd pidfile
@@ -199,6 +219,14 @@ writepid(struct nsd *nsd)
 	return 0;
 }
 
+void
+unlinkpid(const char* file)
+{
+	if (file && unlink(file) == -1)
+		log_msg(LOG_ERR, "failed to unlink pidfile %s: %s",
+			file, strerror(errno));
+}
+
 /*
  * Incoming signals, set appropriate actions.
  *
@@ -216,7 +244,6 @@ sig_handler(int sig)
 			break;
 		case SIGALRM:
 			break;
-		case SIGHUP:
 		case SIGINT:
 		case SIGTERM:
 			nsd.signal_hint_quit = 1;
@@ -338,6 +365,10 @@ main(int argc, char *argv[])
 	pid_t	oldpid;
 	size_t i;
 	struct sigaction action;
+	FILE* dbfd;
+#ifdef HAVE_GETPWNAM
+	struct passwd *pwd;
+#endif /* HAVE_GETPWNAM */
 
 	/* For initialising the address info structures */
 	struct addrinfo hints[MAX_INTERFACES];
@@ -347,6 +378,8 @@ main(int argc, char *argv[])
 
 	const char *configfile = CONFIGFILE;
 
+	char* argv0 = (argv0 = strrchr(argv[0], '/')) ? argv0 + 1 : argv[0];
+
 	log_init("nsd");
 
 	/* Initialize the server handler... */
@@ -376,16 +409,6 @@ main(int argc, char *argv[])
 	nsd.grab_ip6_optional = 0;
 	nsd.file_rotation_ok = 0;
 
-	/* EDNS0 */
-	edns_init_data(&nsd.edns_ipv4, EDNS_MAX_MESSAGE_LEN);
-#if defined(INET6)
-#if defined(IPV6_USE_MIN_MTU)
-	edns_init_data(&nsd.edns_ipv6, EDNS_MAX_MESSAGE_LEN);
-#else /* !defined(IPV6_USE_MIN_MTU) */
-	edns_init_data(&nsd.edns_ipv6, IPV6_MIN_MTU);
-#endif /* defined(IPV6_USE_MIN_MTU) */
-#endif /* defined(INET6) */
-
 	/* Set up our default identity to gethostname(2) */
 	if (gethostname(hostname, MAXHOSTNAMELEN) == 0) {
 		nsd.identity = hostname;
@@ -604,6 +627,11 @@ main(int argc, char *argv[])
 	if(nsd.maximum_tcp_count == 0) {
 		nsd.maximum_tcp_count = nsd.options->tcp_count;
 	}
+	nsd.tcp_timeout = nsd.options->tcp_timeout;
+	nsd.tcp_query_count = nsd.options->tcp_query_count;
+	nsd.ipv4_edns_size = nsd.options->ipv4_edns_size;
+	nsd.ipv6_edns_size = nsd.options->ipv6_edns_size;
+
 	if(udp_port == 0)
 	{
 		if(nsd.options->port != 0) {
@@ -635,6 +663,21 @@ main(int argc, char *argv[])
 			nsd.options->zonesdir));
 	}
 
+	/* EDNS0 */
+	edns_init_data(&nsd.edns_ipv4, nsd.options->ipv4_edns_size);
+#if defined(INET6)
+#if defined(IPV6_USE_MIN_MTU) || defined(IPV6_MTU)
+	edns_init_data(&nsd.edns_ipv6, nsd.options->ipv6_edns_size);
+#else /* no way to set IPV6 MTU, send no bigger than that. */
+	if (nsd.options->ipv6_edns_size < IPV6_MIN_MTU)
+		edns_init_data(&nsd.edns_ipv6, nsd.options->ipv6_edns_size);
+	else
+		edns_init_data(&nsd.edns_ipv6, IPV6_MIN_MTU);
+#endif /* IPV6 MTU) */
+#endif /* defined(INET6) */
+
+
+
 #ifdef NSID
 	edns_init_nsid(&nsd.edns_ipv4, nsd.nsid_len);
 #if defined(INET6)
@@ -725,8 +768,9 @@ main(int argc, char *argv[])
 	/* Parse the username into uid and gid */
 	nsd.gid = getgid();
 	nsd.uid = getuid();
+#ifdef HAVE_GETPWNAM
+	/* Parse the username into uid and gid */
 	if (*nsd.username) {
-		struct passwd *pwd;
 		if (isdigit((int)*nsd.username)) {
 			char *t;
 			nsd.uid = strtol(nsd.username, &t, 10);
@@ -742,7 +786,6 @@ main(int argc, char *argv[])
 				} else {
 					nsd.gid = pwd->pw_gid;
 				}
-				endpwent();
 			}
 		} else {
 			/* Lookup the user id in /etc/passwd */
@@ -752,9 +795,10 @@ main(int argc, char *argv[])
 				nsd.uid = pwd->pw_uid;
 				nsd.gid = pwd->pw_gid;
 			}
-			endpwent();
 		}
 	}
+	/* endpwent(); */
+#endif /* HAVE_GETPWNAM */
 
 #ifdef TSIG
 	if(!tsig_init(nsd.region))
@@ -810,8 +854,8 @@ main(int argc, char *argv[])
 	} else {
 		if (kill(oldpid, 0) == 0 || errno == EPERM) {
 			log_msg(LOG_WARNING,
-				"nsd is already running as %u, continuing",
-				(unsigned) oldpid);
+				"%s is already running as %u, continuing",
+				argv0, (unsigned) oldpid);
 		} else {
 			log_msg(LOG_ERR,
 				"...stale pid file from process %u",
@@ -878,15 +922,115 @@ main(int argc, char *argv[])
 	nsd.signal_hint_statsusr = 0;
 	nsd.quit_sync_done = 0;
 
-	/* Run the server... */
+	/* Initialize the server... */
 	if (server_init(&nsd) != 0) {
-		log_msg(LOG_ERR, "server initialization failed, nsd could "
-						 "not be started");
+		log_msg(LOG_ERR, "server initialization failed, %s could "
+			"not be started", argv0);
 		exit(1);
 	}
 
-	log_msg(LOG_NOTICE, "nsd started (%s), pid %d", PACKAGE_STRING,
-		(int) nsd.pid);
+#ifdef HAVE_CHROOT
+	/* Chroot */
+	if (nsd.chrootdir && strlen(nsd.chrootdir)) {
+		int l = strlen(nsd.chrootdir);
+		int ret = 0;
+
+		while (l>0 && nsd.chrootdir[l-1] == '/')
+			--l;
+
+		/* filename after chroot */
+		ret = file_inside_chroot(nsd.log_filename, nsd.chrootdir);
+		if (ret) {
+			nsd.file_rotation_ok = 1;
+			if (ret == 2) /* also strip chroot */
+				nsd.log_filename += l;
+		}
+		nsd.dbfile += l;
+		nsd.pidfile += l;
+		nsd.options->xfrdfile += l;
+		nsd.options->difffile += l;
+
+		if (chroot(nsd.chrootdir)) {
+			log_msg(LOG_ERR, "unable to chroot: %s", strerror(errno));
+			exit(1);
+		}
+		DEBUG(DEBUG_IPC,1, (LOG_INFO, "changed root directory to %s",
+			nsd.chrootdir));
+	}
+	else
+#endif /* HAVE_CHROOT */
+		nsd.file_rotation_ok = 1;
+
+	DEBUG(DEBUG_IPC,1, (LOG_INFO, "file rotation on %s %sabled",
+		nsd.log_filename, nsd.file_rotation_ok?"en":"dis"));
+
+	/* Check if nsd.db exists */
+	if ((dbfd = fopen(nsd.dbfile, "r")) == NULL) {
+		log_msg(LOG_ERR, "unable to open %s for reading: %s", nsd.dbfile, strerror(errno));
+		exit(1);
+	}
+	fclose(dbfd);
+
+	/* Write pidfile */
+	if (writepid(&nsd) == -1) {
+		log_msg(LOG_ERR, "cannot overwrite the pidfile %s: %s",
+			nsd.pidfile, strerror(errno));
+	}
+
+	/* Drop the permissions */
+#ifdef HAVE_GETPWNAM
+	if (*nsd.username) {
+#ifdef HAVE_SETUSERCONTEXT
+		/* setusercontext does initgroups, setuid, setgid, and
+		 * also resource limits from login config, but we
+		 * still call setresuid, setresgid to be sure to set all uid */
+		if (setusercontext(NULL, pwd, nsd.uid, LOGIN_SETALL) != 0)
+			log_msg(LOG_WARNING, "unable to setusercontext %s: %s",
+				nsd.username, strerror(errno));
+#else /* !HAVE_SETUSERCONTEXT */
+ #ifdef HAVE_INITGROUPS
+		if(initgroups(nsd.username, nsd.gid) != 0)
+			log_msg(LOG_WARNING, "unable to initgroups %s: %s",
+				nsd.username, strerror(errno));
+ #endif /* HAVE_INITGROUPS */
+#endif /* HAVE_SETUSERCONTEXT */
+		endpwent();
+
+#ifdef HAVE_SETRESGID
+		if(setresgid(nsd.gid,nsd.gid,nsd.gid) != 0)
+#elif defined(HAVE_SETREGID) && !defined(DARWIN_BROKEN_SETREUID)
+			if(setregid(nsd.gid,nsd.gid) != 0)
+#else /* use setgid */
+				if(setgid(nsd.gid) != 0)
+#endif /* HAVE_SETRESGID */
+					error("unable to set group id of %s: %s",
+						nsd.username, strerror(errno));
+
+#ifdef HAVE_SETRESUID
+		if(setresuid(nsd.uid,nsd.uid,nsd.uid) != 0)
+#elif defined(HAVE_SETREUID) && !defined(DARWIN_BROKEN_SETREUID)
+			if(setreuid(nsd.uid,nsd.uid) != 0)
+#else /* use setuid */
+				if(setuid(nsd.uid) != 0)
+#endif /* HAVE_SETRESUID */
+					error("unable to set user id of %s: %s",
+						nsd.username, strerror(errno));
+
+		DEBUG(DEBUG_IPC,1, (LOG_INFO, "dropped user privileges, run as %s",
+			nsd.username));
+    }
+#endif /* HAVE_GETPWNAM */
+
+	if (server_prepare(&nsd) != 0) {
+		log_msg(LOG_ERR, "server preparation failed, %s could "
+			"not be started", argv0);
+		unlinkpid(nsd.pidfile);
+		exit(1);
+	}
+
+	/* Really take off */
+	log_msg(LOG_NOTICE, "%s started (%s), pid %d",
+		argv0, PACKAGE_STRING, (int) nsd.pid);
 
 	if (nsd.server_kind == NSD_SERVER_MAIN) {
 		server_main(&nsd);
