@@ -38,6 +38,13 @@
 #include "nsec3.h"
 #include "tsig.h"
 
+/* [Bug #253] Adding unnecessary NS RRset may lead to undesired truncation.
+ * This function determines if the final response packet needs the NS RRset included.
+ * Currently, it will only return negative if QTYPE == DNSKEY. This way, resolvers
+ * won't fallback to TCP unnecessarily when priming DNSKEYs. 
+ */
+static int answer_needs_ns(struct query  *query);
+
 static int add_rrset(struct query  *query,
 		     answer_type    *answer,
 		     rr_section_type section,
@@ -298,7 +305,7 @@ process_query_section(query_type *query)
  * Return NSD_RC_FORMAT on failure, NSD_RC_OK on success.
  */
 static nsd_rc_type
-process_edns(struct query *q)
+process_edns(nsd_type* nsd, struct query *q)
 {
 	if (q->edns.status == EDNS_ERROR) {
 		return NSD_RC_FORMAT;
@@ -307,13 +314,21 @@ process_edns(struct query *q)
 	if (q->edns.status == EDNS_OK) {
 		/* Only care about UDP size larger than normal... */
 		if (!q->tcp && q->edns.maxlen > UDP_MAX_MESSAGE_LEN) {
-			if (q->edns.maxlen < EDNS_MAX_MESSAGE_LEN) {
+			size_t edns_size;
+#if defined(INET6)
+			if (q->addr.ss_family == AF_INET6) {
+				edns_size = nsd->ipv6_edns_size;
+			} else
+#endif
+			edns_size = nsd->ipv4_edns_size;
+
+			if (q->edns.maxlen < edns_size) {
 				q->maxlen = q->edns.maxlen;
 			} else {
-				q->maxlen = EDNS_MAX_MESSAGE_LEN;
+				q->maxlen = edns_size;
 			}
 
-#if defined(INET6) && !defined(IPV6_USE_MIN_MTU)
+#if defined(INET6) && !defined(IPV6_USE_MIN_MTU) && !defined(IPV6_MTU)
 			/*
 			 * Use IPv6 minimum MTU to avoid sending
 			 * packets that are too large for some links.
@@ -606,6 +621,14 @@ add_additional_rrsets(struct query *query, answer_type *answer,
 	}
 }
 
+static int
+answer_needs_ns(struct query* query)
+{
+	assert(query);
+	/* Currently, only troublesome for DNSKEYs, cuz their RRSETs are quite large. */
+	return (query->qtype != TYPE_DNSKEY);
+}
+
 static int
 add_rrset(struct query   *query,
 	  answer_type    *answer,
@@ -902,7 +925,7 @@ answer_domain(struct nsd* nsd, struct query *q, answer_type *answer,
 
 	q->domain = domain;
 
-	if (q->qclass != CLASS_ANY && q->zone->ns_rrset) {
+	if (q->qclass != CLASS_ANY && q->zone->ns_rrset && answer_needs_ns(q)) {
 		add_rrset(q, answer, AUTHORITY_SECTION, q->zone->apex,
 			  q->zone->ns_rrset);
 	}
@@ -1311,7 +1334,7 @@ query_process(query_type *q, nsd_type *nsd)
 		return query_error(q, rc);
 	}
 #endif /* TSIG */
-	rc = process_edns(q);
+	rc = process_edns(nsd, q);
 	if (rc != NSD_RC_OK) {
 		/* We should not return FORMERR, but BADVERS (=16).
 		 * BADVERS is created with Ext. RCODE, followed by RCODE.
