@@ -80,32 +80,32 @@ struct tcp_handler_data
 	 * data, including this structure.  This region is destroyed
 	 * when the connection is closed.
 	 */
-	region_type     *region;
+	region_type*		region;
 
 	/*
 	 * The global nsd structure.
 	 */
-	struct nsd      *nsd;
+	struct nsd*			nsd;
 
 	/*
 	 * The current query data for this TCP connection.
 	 */
-	query_type      *query;
+	query_type*			query;
 
 	/*
 	 * These fields are used to enable the TCP accept handlers
 	 * when the number of TCP connection drops below the maximum
 	 * number of TCP connections.
 	 */
-	size_t              tcp_accept_handler_count;
-	netio_handler_type *tcp_accept_handlers;
-	
+	size_t				tcp_accept_handler_count;
+	netio_handler_type*	tcp_accept_handlers;
+
 	/*
 	 * The query_state is used to remember if we are performing an
 	 * AXFR, if we're done processing, or if we should discard the
 	 * query and connection.
 	 */
-	query_state_type query_state;
+	query_state_type	query_state;
 
 	/*
 	 * The bytes_transmitted field is used to remember the number
@@ -113,7 +113,12 @@ struct tcp_handler_data
 	 * packet.  The count includes the two additional bytes used
 	 * to specify the packet length on a TCP connection.
 	 */
-	size_t           bytes_transmitted;
+	size_t				bytes_transmitted;
+
+	/*
+	 * The number of queries handled by this specific TCP connection.
+	 */
+	int					query_count;
 };
 
 /*
@@ -160,7 +165,9 @@ static void handle_tcp_writing(netio_type *netio,
 static void send_children_quit(struct nsd* nsd);
 
 /* set childrens flags to send NSD_STATS to them */
+#ifdef BIND8_STATS
 static void set_children_stats(struct nsd* nsd);
+#endif /* BIND8_STATS */
 
 /*
  * Change the event types the HANDLERS are interested in to
@@ -192,9 +199,11 @@ delete_child_pid(struct nsd *nsd, pid_t pid)
 		if (nsd->children[i].pid == pid) {
 			nsd->children[i].pid = 0;
 			if(!nsd->children[i].need_to_exit) {
-				if(nsd->children[i].child_fd > 0) close(nsd->children[i].child_fd);
+				if(nsd->children[i].child_fd > 0)
+					close(nsd->children[i].child_fd);
 				nsd->children[i].child_fd = -1;
-				if(nsd->children[i].handler) nsd->children[i].handler->fd = -1;
+				if(nsd->children[i].handler)
+					nsd->children[i].handler->fd = -1;
 			}
 			return i;
 		}
@@ -300,7 +309,7 @@ static void set_bind8_alarm(struct nsd* nsd)
 #endif
 
 static void
-cleanup_dname_compression_tables(void *ptr) 
+cleanup_dname_compression_tables(void *ptr)
 {
 	free(ptr);
 	compressed_dname_offsets = NULL;
@@ -315,7 +324,7 @@ initialize_dname_compression_tables(struct nsd *nsd)
 	if(compression_table_capacity < needed) {
 		compressed_dname_offsets = (uint16_t *) xalloc(
 			needed * sizeof(uint16_t));
-		region_add_cleanup(nsd->db->region, cleanup_dname_compression_tables, 
+		region_add_cleanup(nsd->db->region, cleanup_dname_compression_tables,
 			compressed_dname_offsets);
 		compression_table_capacity = needed;
 		compression_table_size=domain_table_count(nsd->db->domains)+1;
@@ -326,14 +335,13 @@ initialize_dname_compression_tables(struct nsd *nsd)
 
 /*
  * Initialize the server, create and bind the sockets.
- * Drop the priviledges and chroot if requested.
  *
  */
 int
 server_init(struct nsd *nsd)
 {
 	size_t i;
-#if defined(SO_REUSEADDR) || (defined(INET6) && (defined(IPV6_V6ONLY) || defined(IPV6_USE_MIN_MTU)))
+#if defined(SO_REUSEADDR) || (defined(INET6) && (defined(IPV6_V6ONLY) || defined(IPV6_USE_MIN_MTU) || defined(IPV6_MTU)))
 	int on = 1;
 #endif
 
@@ -386,8 +394,45 @@ server_init(struct nsd *nsd)
 					strerror(errno));
 				return -1;
 			}
+# elif defined(IPV6_MTU)
+			/*
+			 * On Linux, PMTUD is disabled by default for datagrams
+			 * so set the MTU equal to the MIN MTU to get the same.
+			 */
+			on = IPV6_MIN_MTU;
+			if (setsockopt(nsd->udp[i].s, IPPROTO_IPV6, IPV6_MTU, 
+				&on, sizeof(on)) < 0)
+			{
+				log_msg(LOG_ERR, "setsockopt(..., IPV6_MTU, ...) failed: %s",
+					strerror(errno));
+				return -1;
+			}
+			on = 1;
 # endif
 		}
+#endif
+#if defined(AF_INET)
+		if (nsd->udp[i].addr->ai_family == AF_INET) {
+#  if defined(IP_MTU_DISCOVER) && defined(IP_PMTUDISC_DONT)
+			int action = IP_PMTUDISC_DONT;
+			if (setsockopt(nsd->udp[i].s, IPPROTO_IP, 
+				IP_MTU_DISCOVER, &action, sizeof(action)) < 0)
+			{
+				log_msg(LOG_ERR, "setsockopt(..., IP_MTU_DISCOVER, IP_PMTUDISC_DONT...) failed: %s",
+					strerror(errno));
+				return -1;
+			}
+#  elif defined(IP_DONTFRAG)
+			int off = 0;
+			if (setsockopt(nsd->udp[i].s, IPPROTO_IP, IP_DONTFRAG,
+				&off, sizeof(off)) < 0)
+			{
+				log_msg(LOG_ERR, "setsockopt(..., IP_DONTFRAG, ...) failed: %s",
+					strerror(errno));
+				return -1;
+			}
+#  endif
+		}
 #endif
 		/* set it nonblocking */
 		/* otherwise, on OSes with thundering herd problems, the
@@ -398,7 +443,7 @@ server_init(struct nsd *nsd)
 
 		/* Bind it... */
 		if (bind(nsd->udp[i].s, (struct sockaddr *) nsd->udp[i].addr->ai_addr, nsd->udp[i].addr->ai_addrlen) != 0) {
-			log_msg(LOG_ERR, "can't bind the socket: %s", strerror(errno));
+			log_msg(LOG_ERR, "can't bind udp socket: %s", strerror(errno));
 			return -1;
 		}
 	}
@@ -446,7 +491,7 @@ server_init(struct nsd *nsd)
 
 		/* Bind it... */
 		if (bind(nsd->tcp[i].s, (struct sockaddr *) nsd->tcp[i].addr->ai_addr, nsd->tcp[i].addr->ai_addrlen) != 0) {
-			log_msg(LOG_ERR, "can't bind the socket: %s", strerror(errno));
+			log_msg(LOG_ERR, "can't bind tcp socket: %s", strerror(errno));
 			return -1;
 		}
 
@@ -457,44 +502,36 @@ server_init(struct nsd *nsd)
 		}
 	}
 
-#ifdef HAVE_CHROOT
-	/* Chroot */
-	if (nsd->chrootdir) {
-		int l = strlen(nsd->chrootdir);
-
-		nsd->dbfile += l;
-		nsd->pidfile += l;
-		nsd->options->xfrdfile += l;
-		nsd->options->difffile += l;
-
-		if (chroot(nsd->chrootdir)) {
-			log_msg(LOG_ERR, "unable to chroot: %s", strerror(errno));
-			return -1;
-		}
-	}
-#endif
-
-	/* Drop the permissions */
-	if (setgid(nsd->gid) != 0 || setuid(nsd->uid) !=0) {
-		log_msg(LOG_ERR, "unable to drop user priviledges: %s", strerror(errno));
-		return -1;
-	}
+	return 0;
+}
 
+/*
+ * Prepare the server for take off.
+ *
+ */
+int
+server_prepare(struct nsd *nsd)
+{
 	/* Open the database... */
 	if ((nsd->db = namedb_open(nsd->dbfile, nsd->options, nsd->child_count)) == NULL) {
+		log_msg(LOG_ERR, "unable to open the database %s: %s",
+			nsd->dbfile, strerror(errno));
 		return -1;
 	}
-	if(!diff_read_file(nsd->db, nsd->options, NULL, nsd->child_count))
-	{
-		log_msg(LOG_ERR, "The diff file contains errors. Will continue without it");
+
+	/* Read diff file */
+	if(!diff_read_file(nsd->db, nsd->options, NULL, nsd->child_count)) {
+		log_msg(LOG_ERR, "The diff file contains errors. Will continue "
+						 "without it");
 	}
+
 #ifdef NSEC3
 	prehash(nsd->db, 0);
 #endif
 
 	compression_table_capacity = 0;
 	initialize_dname_compression_tables(nsd);
-	
+
 #ifdef	BIND8_STATS
 	/* Initialize times... */
 	time(&nsd->st.boot);
@@ -530,6 +567,7 @@ close_all_sockets(struct nsd_socket sockets[], size_t n)
 	for (i = 0; i < n; ++i) {
 		if (sockets[i].s != -1) {
 			close(sockets[i].s);
+			free(sockets[i].addr);
 			sockets[i].s = -1;
 		}
 	}
@@ -556,7 +594,7 @@ server_shutdown(struct nsd *nsd)
 	/* SERVER: close command channels to children */
 	if(!nsd->this_child)
 	{
-		for(i=0; i<nsd->child_count; ++i)
+		for(i=0; i < nsd->child_count; ++i)
 			if(nsd->children[i].child_fd > 0)
 			{
 				close(nsd->children[i].child_fd);
@@ -564,6 +602,12 @@ server_shutdown(struct nsd *nsd)
 			}
 	}
 
+	log_finalize();
+	tsig_finalize();
+
+	nsd_options_destroy(nsd->options);
+	region_destroy(nsd->region);
+
 	exit(0);
 }
 
@@ -579,7 +623,8 @@ server_start_xfrd(struct nsd *nsd, netio_handler_type* handler)
 		zone->updated = 0;
 	}
 
-	if(handler->fd != -1) close(handler->fd);
+	if(handler->fd != -1)
+		close(handler->fd);
 	if (socketpair(AF_UNIX, SOCK_STREAM, 0, sockets) == -1) {
 		log_msg(LOG_ERR, "startxfrd failed on socketpair: %s", strerror(errno));
 		return -1;
@@ -590,17 +635,18 @@ server_start_xfrd(struct nsd *nsd, netio_handler_type* handler)
 		log_msg(LOG_ERR, "fork xfrd failed: %s", strerror(errno));
 		break;
 	case 0:
-		/* CHILD */
+		/* CHILD: close first socket, use second one */
 		close(sockets[0]);
 		xfrd_init(sockets[1], nsd);
 		/* ENOTREACH */
 		break;
 	default:
-		/* PARENT */
+		/* PARENT: close second socket, use first one */
 		close(sockets[1]);
 		handler->fd = sockets[0];
 		break;
 	}
+	/* PARENT only */
 	handler->timeout = NULL;
 	handler->event_types = NETIO_EVENT_READ;
 	handler->event_handler = parent_handle_xfrd_command;
@@ -611,7 +657,7 @@ server_start_xfrd(struct nsd *nsd, netio_handler_type* handler)
 }
 
 /* pass timeout=-1 for blocking. Returns size, 0, -1(err), or -2(timeout) */
-static ssize_t 
+static ssize_t
 block_read(struct nsd* nsd, int s, void* p, ssize_t sz, int timeout)
 {
 	uint8_t* buf = (uint8_t*) p;
@@ -671,7 +717,7 @@ block_read(struct nsd* nsd, int s, void* p, ssize_t sz, int timeout)
  * as server_main.
  */
 static void
-server_reload(struct nsd *nsd, region_type* server_region, netio_type* netio, 
+server_reload(struct nsd *nsd, region_type* server_region, netio_type* netio,
 	int cmdsocket, int* xfrd_sock_p)
 {
 	pid_t old_pid;
@@ -681,34 +727,35 @@ server_reload(struct nsd *nsd, region_type* server_region, netio_type* netio,
 	int ret;
 
 	if(db_crc_different(nsd->db) == 0) {
-		DEBUG(DEBUG_XFRD,1, (LOG_INFO, 
+		DEBUG(DEBUG_XFRD,1, (LOG_INFO,
 			"CRC the same. skipping %s.", nsd->db->filename));
 	} else {
-		DEBUG(DEBUG_XFRD,1, (LOG_INFO, 
+		DEBUG(DEBUG_XFRD,1, (LOG_INFO,
 			"CRC different. reread of %s.", nsd->db->filename));
 		namedb_close(nsd->db);
-		if ((nsd->db = namedb_open(nsd->dbfile, nsd->options, 
+		if ((nsd->db = namedb_open(nsd->dbfile, nsd->options,
 			nsd->child_count)) == NULL) {
 			log_msg(LOG_ERR, "unable to reload the database: %s", strerror(errno));
 			exit(1);
 		}
 	}
 	if(!diff_read_file(nsd->db, nsd->options, NULL, nsd->child_count)) {
-			log_msg(LOG_ERR, "unable to load the diff file: %s", nsd->options->difffile);
-			exit(1);
+		log_msg(LOG_ERR, "unable to load the diff file: %s", nsd->options->difffile);
+		exit(1);
 	}
 	log_msg(LOG_INFO, "memory recyclebin holds %lu bytes", (unsigned long)
 		region_get_recycle_size(nsd->db->region));
 #ifndef NDEBUG
 	if(nsd_debug_level >= 1)
 		region_log_stats(nsd->db->region);
-#endif
+#endif /* NDEBUG */
 #ifdef NSEC3
 	prehash(nsd->db, 1);
-#endif
+#endif /* NSEC3 */
 
 	initialize_dname_compression_tables(nsd);
 
+	/* Get our new process id */
 	old_pid = nsd->pid;
 	nsd->pid = getpid();
 
@@ -718,11 +765,21 @@ server_reload(struct nsd *nsd, region_type* server_region, netio_type* netio,
 	set_bind8_alarm(nsd);
 #endif
 
+	/* Start new child processes */
 	if (server_start_children(nsd, server_region, netio, xfrd_sock_p) != 0) {
 		send_children_quit(nsd);
 		exit(1);
 	}
 
+	/* Overwrite pid before closing old parent, to avoid race condition:
+	 * - parent process already closed
+	 * - pidfile still contains old_pid
+	 * - control script contacts parent process, using contents of pidfile
+	 */
+	if (writepid(nsd) == -1) {
+		log_msg(LOG_ERR, "cannot overwrite the pidfile %s: %s", nsd->pidfile, strerror(errno));
+	}
+
 #define RELOAD_SYNC_TIMEOUT 25 /* seconds */
 	/* Send quit command to parent: blocking, wait for receipt. */
 	do {
@@ -734,7 +791,7 @@ server_reload(struct nsd *nsd, region_type* server_region, netio_type* netio,
 		}
 		/* blocking: wait for parent to really quit. (it sends RELOAD as ack) */
 		DEBUG(DEBUG_IPC,1, (LOG_INFO, "reload: ipc wait for ack main"));
-		ret = block_read(nsd, cmdsocket, &cmd, sizeof(cmd), 
+		ret = block_read(nsd, cmdsocket, &cmd, sizeof(cmd),
 			RELOAD_SYNC_TIMEOUT);
 		if(ret == -2) {
 			DEBUG(DEBUG_IPC, 1, (LOG_ERR, "reload timeout QUITSYNC. retry"));
@@ -747,11 +804,6 @@ server_reload(struct nsd *nsd, region_type* server_region, netio_type* netio,
 	DEBUG(DEBUG_IPC,1, (LOG_INFO, "reload: ipc reply main %d %d", ret, cmd));
 	assert(ret==-1 || ret == 0 || cmd == NSD_RELOAD);
 
-	/* Overwrite pid... */
-	if (writepid(nsd) == -1) {
-		log_msg(LOG_ERR, "cannot overwrite the pidfile %s: %s", nsd->pidfile, strerror(errno));
-	}
-
 	/* inform xfrd of new SOAs */
 	cmd = NSD_SOA_BEGIN;
 	if(!write_socket(xfrd_sock, &cmd,  sizeof(cmd))) {
@@ -779,7 +831,7 @@ server_reload(struct nsd *nsd, region_type* server_region, netio_type* netio,
 		/* use blocking writes */
 		if(!write_socket(xfrd_sock, &cmd,  sizeof(cmd)) ||
 			!write_socket(xfrd_sock, &sz, sizeof(sz)) ||
-			!write_socket(xfrd_sock, domain_dname(zone->apex), 
+			!write_socket(xfrd_sock, domain_dname(zone->apex),
 				dname_total_size(domain_dname(zone->apex))))
 		{
 			log_msg(LOG_ERR, "problems sending soa info from reload %d to xfrd: %s",
@@ -818,7 +870,10 @@ server_reload(struct nsd *nsd, region_type* server_region, netio_type* netio,
 		log_msg(LOG_ERR, "problems sending soa end from reload %d to xfrd: %s",
 			(int)nsd->pid, strerror(errno));
 	}
-	DEBUG(DEBUG_IPC,2, (LOG_INFO, "Reload exiting to become new main"));
+
+	/* try to reopen file */
+	if (nsd->file_rotation_ok)
+		log_reopen(nsd->log_filename, 1);
 	/* exit reload, continue as new server_main */
 }
 
@@ -866,7 +921,7 @@ server_signal_mode(struct nsd *nsd)
 void
 server_main(struct nsd *nsd)
 {
-        region_type *server_region = region_create(xalloc, free);
+	region_type *server_region = region_create(xalloc, free);
 	netio_type *netio = netio_create(server_region);
 	netio_handler_type reload_listener;
 	netio_handler_type xfrd_listener;
@@ -878,26 +933,34 @@ server_main(struct nsd *nsd)
 	pid_t reload_pid = -1;
 	pid_t xfrd_pid = -1;
 	sig_atomic_t mode;
-	
+
+	/* Ensure we are the main process */
 	assert(nsd->server_kind == NSD_SERVER_MAIN);
 
 	xfrd_listener.user_data = (struct ipc_handler_conn_data*)region_alloc(
 		server_region, sizeof(struct ipc_handler_conn_data));
 	xfrd_listener.fd = -1;
 	((struct ipc_handler_conn_data*)xfrd_listener.user_data)->nsd = nsd;
-	((struct ipc_handler_conn_data*)xfrd_listener.user_data)->conn = 
+	((struct ipc_handler_conn_data*)xfrd_listener.user_data)->conn =
 		xfrd_tcp_create(server_region);
+
+	/* Start the XFRD process */
 	xfrd_pid = server_start_xfrd(nsd, &xfrd_listener);
 	netio_add_handler(netio, &xfrd_listener);
+
+	/* Start the child processes that handle incoming queries */
 	if (server_start_children(nsd, server_region, netio, &xfrd_listener.fd) != 0) {
 		send_children_quit(nsd);
 		exit(1);
 	}
 	reload_listener.fd = -1;
+
+	/* This_child MUST be 0, because this is the parent process */
 	assert(nsd->this_child == 0);
 
+	/* Run the server until we get a shutdown signal */
 	while ((mode = nsd->mode) != NSD_SHUTDOWN) {
-
+		/* Did we receive a signal that changes our mode? */
 		if(mode == NSD_RUN) {
 			nsd->mode = mode = server_signal_mode(nsd);
 		}
@@ -915,7 +978,7 @@ server_main(struct nsd *nsd)
 					log_msg(LOG_WARNING,
 					       "server %d died unexpectedly with status %d, restarting",
 					       (int) child_pid, status);
-					restart_child_servers(nsd, server_region, netio, 
+					restart_child_servers(nsd, server_region, netio,
 						&xfrd_listener.fd);
 				} else if (child_pid == reload_pid) {
 					sig_atomic_t cmd = NSD_SOA_END;
@@ -927,11 +990,11 @@ server_main(struct nsd *nsd)
 					reload_listener.fd = -1;
 					reload_listener.event_types = NETIO_EVENT_NONE;
 					/* inform xfrd reload attempt ended */
-					if(!write_socket(xfrd_listener.fd, 
-						&cmd, sizeof(cmd)) == -1) { 
+					if(!write_socket(xfrd_listener.fd,
+						&cmd, sizeof(cmd)) == -1) {
 						log_msg(LOG_ERR, "problems "
-						  "sending SOAEND to xfrd: %s", 
-						  strerror(errno)); 
+						  "sending SOAEND to xfrd: %s",
+						  strerror(errno));
 					}
 				} else if (child_pid == xfrd_pid) {
 					log_msg(LOG_WARNING,
@@ -954,7 +1017,7 @@ server_main(struct nsd *nsd)
 				break;
 
 			/* timeout to collect processes. In case no sigchild happens. */
-			timeout_spec.tv_sec = 60; 
+			timeout_spec.tv_sec = 60;
 			timeout_spec.tv_nsec = 0;
 
 			/* listen on ports, timeout for collecting terminated children */
@@ -966,6 +1029,7 @@ server_main(struct nsd *nsd)
 
 			break;
 		case NSD_RELOAD:
+			/* Continue to run nsd after reload */
 			nsd->mode = NSD_RUN;
 
 			if (reload_pid != -1) {
@@ -981,6 +1045,8 @@ server_main(struct nsd *nsd)
 				reload_pid = -1;
 				break;
 			}
+
+			/* Do actual reload */
 			reload_pid = fork();
 			switch (reload_pid) {
 			case -1:
@@ -989,7 +1055,7 @@ server_main(struct nsd *nsd)
 			case 0:
 				/* CHILD */
 				close(reload_sockets[0]);
-				server_reload(nsd, server_region, netio, 
+				server_reload(nsd, server_region, netio,
 					reload_sockets[1], &xfrd_listener.fd);
 				DEBUG(DEBUG_IPC,2, (LOG_INFO, "Reload exited to become new main"));
 				close(reload_sockets[1]);
@@ -1003,7 +1069,9 @@ server_main(struct nsd *nsd)
 				DEBUG(DEBUG_IPC,2, (LOG_INFO, "Reload resetup; run"));
 				break;
 			default:
-				/* PARENT */
+				/* PARENT, keep running until NSD_QUIT_SYNC
+				 * received from CHILD.
+				 */
 				close(reload_sockets[1]);
 				reload_listener.fd = reload_sockets[0];
 				reload_listener.timeout = NULL;
@@ -1014,12 +1082,12 @@ server_main(struct nsd *nsd)
 				break;
 			}
 			break;
-		case NSD_QUIT_SYNC: 
+		case NSD_QUIT_SYNC:
 			/* synchronisation of xfrd, parent and reload */
 			if(!nsd->quit_sync_done && reload_listener.fd > 0) {
 				sig_atomic_t cmd = NSD_RELOAD;
 				/* stop xfrd ipc writes in progress */
-				DEBUG(DEBUG_IPC,1, (LOG_INFO, 
+				DEBUG(DEBUG_IPC,1, (LOG_INFO,
 					"main: ipc send indication reload"));
 				if(!write_socket(xfrd_listener.fd, &cmd, sizeof(cmd))) {
 					log_msg(LOG_ERR, "server_main: could not send reload "
@@ -1031,7 +1099,7 @@ server_main(struct nsd *nsd)
 			}
 			nsd->mode = NSD_RUN;
 			break;
-		case NSD_QUIT: 
+		case NSD_QUIT:
 			/* silent shutdown during reload */
 			if(reload_listener.fd > 0) {
 				/* acknowledge the quit, to sync reload that we will really quit now */
@@ -1045,8 +1113,11 @@ server_main(struct nsd *nsd)
 			}
 			/* only quit children after xfrd has acked */
 			send_children_quit(nsd);
+
 			region_destroy(server_region);
+			namedb_close(nsd->db);
 			server_shutdown(nsd);
+
 			/* ENOTREACH */
 			break;
 		case NSD_SHUTDOWN:
@@ -1077,21 +1148,24 @@ server_main(struct nsd *nsd)
 	close(fd);
 
 	/* Unlink it if possible... */
-	unlink(nsd->pidfile);
+	unlinkpid(nsd->pidfile);
 
-	if(reload_listener.fd > 0) close(reload_listener.fd);
+	if(reload_listener.fd > 0)
+		close(reload_listener.fd);
 	if(xfrd_listener.fd > 0) {
 		/* complete quit, stop xfrd */
 		sig_atomic_t cmd = NSD_QUIT;
-		DEBUG(DEBUG_IPC,1, (LOG_INFO, 
+		DEBUG(DEBUG_IPC,1, (LOG_INFO,
 			"main: ipc send quit to xfrd"));
 		if(!write_socket(xfrd_listener.fd, &cmd, sizeof(cmd))) {
-			log_msg(LOG_ERR, "server_main: could not send quit to xfrd: %s", 
+			log_msg(LOG_ERR, "server_main: could not send quit to xfrd: %s",
 				strerror(errno));
 		}
 		fsync(xfrd_listener.fd);
 		close(xfrd_listener.fd);
 	}
+
+	namedb_close(nsd->db);
 	region_destroy(server_region);
 	server_shutdown(nsd);
 }
@@ -1113,17 +1187,19 @@ server_child(struct nsd *nsd)
 	region_type *server_region = region_create(xalloc, free);
 	netio_type *netio = netio_create(server_region);
 	netio_handler_type *tcp_accept_handlers;
+	query_type *udp_query;
 	sig_atomic_t mode;
-	
+
 	assert(nsd->server_kind != NSD_SERVER_MAIN);
-	
+	DEBUG(DEBUG_IPC, 2, (LOG_INFO, "child process started"));
+
 	if (!(nsd->server_kind & NSD_SERVER_TCP)) {
 		close_all_sockets(nsd->tcp, nsd->ifs);
 	}
 	if (!(nsd->server_kind & NSD_SERVER_UDP)) {
 		close_all_sockets(nsd->udp, nsd->ifs);
 	}
-	
+
 	if (nsd->this_child && nsd->this_child->parent_fd != -1) {
 		netio_handler_type *handler;
 
@@ -1142,6 +1218,9 @@ server_child(struct nsd *nsd)
 	}
 
 	if (nsd->server_kind & NSD_SERVER_UDP) {
+		udp_query = query_create(server_region,
+			compressed_dname_offsets, compression_table_size);
+
 		for (i = 0; i < nsd->ifs; ++i) {
 			struct udp_handler_data *data;
 			netio_handler_type *handler;
@@ -1149,9 +1228,7 @@ server_child(struct nsd *nsd)
 			data = (struct udp_handler_data *) region_alloc(
 				server_region,
 				sizeof(struct udp_handler_data));
-			data->query = query_create(
-				server_region, compressed_dname_offsets, 
-				compression_table_size);
+			data->query = udp_query;
 			data->nsd = nsd;
 			data->socket = &nsd->udp[i];
 
@@ -1177,7 +1254,7 @@ server_child(struct nsd *nsd)
 		for (i = 0; i < nsd->ifs; ++i) {
 			struct tcp_accept_handler_data *data;
 			netio_handler_type *handler;
-			
+
 			data = (struct tcp_accept_handler_data *) region_alloc(
 				server_region,
 				sizeof(struct tcp_accept_handler_data));
@@ -1185,7 +1262,7 @@ server_child(struct nsd *nsd)
 			data->socket = &nsd->tcp[i];
 			data->tcp_accept_handler_count = nsd->ifs;
 			data->tcp_accept_handlers = tcp_accept_handlers;
-			
+
 			handler = &tcp_accept_handlers[i];
 			handler->fd = nsd->tcp[i].s;
 			handler->timeout = NULL;
@@ -1195,8 +1272,8 @@ server_child(struct nsd *nsd)
 			netio_add_handler(netio, handler);
 		}
 	}
-	
-	/* The main loop... */	
+
+	/* The main loop... */
 	while ((mode = nsd->mode) != NSD_QUIT) {
 		if(mode == NSD_RUN) nsd->mode = mode = server_signal_mode(nsd);
 
@@ -1205,7 +1282,7 @@ server_child(struct nsd *nsd)
 #ifdef BIND8_STATS
 			/* Dump the statistics */
 			bind8_stats(nsd);
-#else /* BIND8_STATS */
+#else /* !BIND8_STATS */
 			log_msg(LOG_NOTICE, "Statistics support not enabled at compile time.");
 #endif /* BIND8_STATS */
 
@@ -1216,13 +1293,14 @@ server_child(struct nsd *nsd)
 			if (nsd->this_child->parent_fd > 0) {
 				sig_atomic_t parent_notify = NSD_REAP_CHILDREN;
 				if (write(nsd->this_child->parent_fd,
-				    &parent_notify, 
+				    &parent_notify,
 				    sizeof(parent_notify)) == -1)
 				{
 					log_msg(LOG_ERR, "problems sending command from %d to parent: %s",
 						(int) nsd->this_child->pid, strerror(errno));
 				}
-			} else while (waitpid(0, NULL, WNOHANG) > 0) /* no parent, so reap 'em */;
+			} else /* no parent, so reap 'em */
+				while (waitpid(0, NULL, WNOHANG) > 0) ;
 			nsd->mode = NSD_RUN;
 		}
 		else if(mode == NSD_RUN) {
@@ -1236,7 +1314,7 @@ server_child(struct nsd *nsd)
 		} else if(mode == NSD_QUIT) {
 			/* ignore here, quit */
 		} else {
-			log_msg(LOG_ERR, "mode bad value %d, back to service.", 
+			log_msg(LOG_ERR, "mode bad value %d, back to service.",
 				mode);
 			nsd->mode = NSD_RUN;
 		}
@@ -1246,8 +1324,8 @@ server_child(struct nsd *nsd)
 	bind8_stats(nsd);
 #endif /* BIND8_STATS */
 
+	namedb_close(nsd->db);
 	region_destroy(server_region);
-	
 	server_shutdown(nsd);
 }
 
@@ -1261,11 +1339,11 @@ handle_udp(netio_type *ATTR_UNUSED(netio),
 		= (struct udp_handler_data *) handler->user_data;
 	int received, sent;
 	struct query *q = data->query;
-	
+
 	if (!(event_types & NETIO_EVENT_READ)) {
 		return;
 	}
-	
+
 	/* Account... */
 	if (data->socket->addr->ai_family == AF_INET) {
 		STATUP(data->nsd, qudp);
@@ -1293,8 +1371,9 @@ handle_udp(netio_type *ATTR_UNUSED(netio),
 
 		/* Process and answer the query... */
 		if (server_process_query(data->nsd, q) != QUERY_DISCARDED) {
-			if (RCODE(q->packet) == RCODE_OK && !AA(q->packet))
+			if (RCODE(q->packet) == RCODE_OK && !AA(q->packet)) {
 				STATUP(data->nsd, nona);
+			}
 
 			/* Add EDNS0 and TSIG info if necessary.  */
 			query_add_optional(q, data->nsd);
@@ -1366,6 +1445,13 @@ handle_tcp_reading(netio_type *netio,
 		return;
 	}
 
+	if (data->nsd->tcp_query_count > 0 &&
+		data->query_count >= data->nsd->tcp_query_count) {
+		/* No more queries allowed on this tcp connection.  */
+		cleanup_tcp_handler(netio, handler);
+		return;
+	}
+
 	assert(event_types & NETIO_EVENT_READ);
 
 	if (data->bytes_transmitted == 0) {
@@ -1388,6 +1474,9 @@ handle_tcp_reading(netio_type *netio,
 				 */
 				return;
 			} else {
+#ifdef ECONNRESET
+				if (verbosity >= 2 || errno != ECONNRESET)
+#endif /* ECONNRESET */
 				log_msg(LOG_ERR, "failed reading from tcp: %s", strerror(errno));
 				cleanup_tcp_handler(netio, handler);
 				return;
@@ -1410,7 +1499,7 @@ handle_tcp_reading(netio_type *netio,
 		assert(data->bytes_transmitted == sizeof(uint16_t));
 
 		data->query->tcplen = ntohs(data->query->tcplen);
-		
+
 		/*
 		 * Minimum query size is:
 		 *
@@ -1448,6 +1537,9 @@ handle_tcp_reading(netio_type *netio,
 			 */
 			return;
 		} else {
+#ifdef ECONNRESET
+			if (verbosity >= 2 || errno != ECONNRESET)
+#endif /* ECONNRESET */
 			log_msg(LOG_ERR, "failed reading from tcp: %s", strerror(errno));
 			cleanup_tcp_handler(netio, handler);
 			return;
@@ -1483,6 +1575,9 @@ handle_tcp_reading(netio_type *netio,
 
 	/* We have a complete query, process it.  */
 
+	/* tcp-query-count: handle query counter ++ */
+	data->query_count++;
+
 	buffer_flip(data->query->packet);
 	data->query_state = server_process_query(data->nsd, data->query);
 	if (data->query_state == QUERY_DISCARDED) {
@@ -1497,18 +1592,18 @@ handle_tcp_reading(netio_type *netio,
 	{
 		STATUP(data->nsd, nona);
 	}
-		
+
 	query_add_optional(data->query, data->nsd);
 
 	/* Switch to the tcp write handler.  */
 	buffer_flip(data->query->packet);
 	data->query->tcplen = buffer_remaining(data->query->packet);
 	data->bytes_transmitted = 0;
-	
-	handler->timeout->tv_sec = TCP_TIMEOUT;
+
+	handler->timeout->tv_sec = data->nsd->tcp_timeout;
 	handler->timeout->tv_nsec = 0L;
 	timespec_add(handler->timeout, netio_current_time(netio));
-	
+
 	handler->event_types = NETIO_EVENT_WRITE | NETIO_EVENT_TIMEOUT;
 	handler->event_handler = handle_tcp_writing;
 }
@@ -1546,8 +1641,8 @@ handle_tcp_writing(netio_type *netio,
 				return;
 			} else {
 #ifdef ECONNRESET
-				if(verbosity>=2 || errno != ECONNRESET)
-#endif
+				if(verbosity >= 2 || errno != ECONNRESET)
+#endif /* ECONNRESET */
 				log_msg(LOG_ERR, "failed writing to tcp: %s", strerror(errno));
 				cleanup_tcp_handler(netio, handler);
 				return;
@@ -1580,8 +1675,8 @@ handle_tcp_writing(netio_type *netio,
 			return;
 		} else {
 #ifdef ECONNRESET
-			if(verbosity>=2 || errno != ECONNRESET)
-#endif
+			if(verbosity >= 2 || errno != ECONNRESET)
+#endif /* ECONNRESET */
 			log_msg(LOG_ERR, "failed writing to tcp: %s", strerror(errno));
 			cleanup_tcp_handler(netio, handler);
 			return;
@@ -1606,13 +1701,13 @@ handle_tcp_writing(netio_type *netio,
 		data->query_state = query_axfr(data->nsd, q);
 		if (data->query_state != QUERY_PROCESSED) {
 			query_add_optional(data->query, data->nsd);
-			
+
 			/* Reset data. */
 			buffer_flip(q->packet);
 			q->tcplen = buffer_remaining(q->packet);
 			data->bytes_transmitted = 0;
 			/* Reset timeout.  */
-			handler->timeout->tv_sec = TCP_TIMEOUT;
+			handler->timeout->tv_sec = data->nsd->tcp_timeout;
 			handler->timeout->tv_nsec = 0;
 			timespec_add(handler->timeout, netio_current_time(netio));
 
@@ -1628,9 +1723,15 @@ handle_tcp_writing(netio_type *netio,
 	 * Done sending, wait for the next request to arrive on the
 	 * TCP socket by installing the TCP read handler.
 	 */
+	if (data->nsd->tcp_query_count > 0 &&
+		data->query_count >= data->nsd->tcp_query_count) {
+
+		(void) shutdown(handler->fd, SHUT_WR);
+	}
+
 	data->bytes_transmitted = 0;
-	
-	handler->timeout->tv_sec = TCP_TIMEOUT;
+
+	handler->timeout->tv_sec = data->nsd->tcp_timeout;
 	handler->timeout->tv_nsec = 0;
 	timespec_add(handler->timeout, netio_current_time(netio));
 
@@ -1661,7 +1762,7 @@ handle_tcp_accept(netio_type *netio,
 	struct sockaddr_in addr;
 #endif
 	socklen_t addrlen;
-	
+
 	if (!(event_types & NETIO_EVENT_READ)) {
 		return;
 	}
@@ -1669,17 +1770,17 @@ handle_tcp_accept(netio_type *netio,
 	if (data->nsd->current_tcp_count >= data->nsd->maximum_tcp_count) {
 		return;
 	}
-	
+
 	/* Accept it... */
 	addrlen = sizeof(addr);
 	s = accept(handler->fd, (struct sockaddr *) &addr, &addrlen);
 	if (s == -1) {
 		/* EINTR is a signal interrupt. The others are various OS ways
 		   of saying that the client has closed the connection. */
-		if (	errno != EINTR 
-			&& errno != EWOULDBLOCK 
+		if (	errno != EINTR
+			&& errno != EWOULDBLOCK
 #ifdef ECONNABORTED
-			&& errno != ECONNABORTED 
+			&& errno != ECONNABORTED
 #endif /* ECONNABORTED */
 #ifdef EPROTO
 			&& errno != EPROTO
@@ -1695,7 +1796,7 @@ handle_tcp_accept(netio_type *netio,
 		close(s);
 		return;
 	}
-	
+
 	/*
 	 * This region is deallocated when the TCP connection is
 	 * closed by the TCP handler.
@@ -1707,7 +1808,8 @@ handle_tcp_accept(netio_type *netio,
 	tcp_data->query = query_create(tcp_region, compressed_dname_offsets,
 		compression_table_size);
 	tcp_data->nsd = data->nsd;
-	
+	tcp_data->query_count = 0;
+
 	tcp_data->tcp_accept_handler_count = data->tcp_accept_handler_count;
 	tcp_data->tcp_accept_handlers = data->tcp_accept_handlers;
 
@@ -1715,13 +1817,13 @@ handle_tcp_accept(netio_type *netio,
 	tcp_data->bytes_transmitted = 0;
 	memcpy(&tcp_data->query->addr, &addr, addrlen);
 	tcp_data->query->addrlen = addrlen;
-	
+
 	tcp_handler = (netio_handler_type *) region_alloc(
 		tcp_region, sizeof(netio_handler_type));
 	tcp_handler->fd = s;
 	tcp_handler->timeout = (struct timespec *) region_alloc(
 		tcp_region, sizeof(struct timespec));
-	tcp_handler->timeout->tv_sec = TCP_TIMEOUT;
+	tcp_handler->timeout->tv_sec = data->nsd->tcp_timeout;
 	tcp_handler->timeout->tv_nsec = 0L;
 	timespec_add(tcp_handler->timeout, netio_current_time(netio));
 
@@ -1744,7 +1846,7 @@ handle_tcp_accept(netio_type *netio,
 	}
 }
 
-static void 
+static void
 send_children_quit(struct nsd* nsd)
 {
 	sig_atomic_t command = NSD_QUIT;
@@ -1769,7 +1871,8 @@ send_children_quit(struct nsd* nsd)
 	}
 }
 
-static void 
+#ifdef BIND8_STATS
+static void
 set_children_stats(struct nsd* nsd)
 {
 	size_t i;
@@ -1780,6 +1883,7 @@ set_children_stats(struct nsd* nsd)
 		nsd->children[i].handler->event_types |= NETIO_EVENT_WRITE;
 	}
 }
+#endif /* BIND8_STATS */
 
 static void
 configure_handler_event_types(size_t count,
@@ -1789,7 +1893,7 @@ configure_handler_event_types(size_t count,
 	size_t i;
 
 	assert(handlers);
-	
+
 	for (i = 0; i < count; ++i) {
 		handlers[i].event_types = event_types;
 	}
