@@ -92,26 +92,27 @@
      T_MAILB = 308,
      T_MAILA = 309,
      T_DS = 310,
-     T_SSHFP = 311,
-     T_RRSIG = 312,
-     T_NSEC = 313,
-     T_DNSKEY = 314,
-     T_SPF = 315,
-     T_NSEC3 = 316,
-     T_IPSECKEY = 317,
-     T_DHCID = 318,
-     T_NSEC3PARAM = 319,
-     DOLLAR_TTL = 320,
-     DOLLAR_ORIGIN = 321,
-     NL = 322,
-     SP = 323,
-     STR = 324,
-     PREV = 325,
-     BITLAB = 326,
-     T_TTL = 327,
-     T_RRCLASS = 328,
-     URR = 329,
-     T_UTYPE = 330
+     T_DLV = 311,
+     T_SSHFP = 312,
+     T_RRSIG = 313,
+     T_NSEC = 314,
+     T_DNSKEY = 315,
+     T_SPF = 316,
+     T_NSEC3 = 317,
+     T_IPSECKEY = 318,
+     T_DHCID = 319,
+     T_NSEC3PARAM = 320,
+     DOLLAR_TTL = 321,
+     DOLLAR_ORIGIN = 322,
+     NL = 323,
+     SP = 324,
+     STR = 325,
+     PREV = 326,
+     BITLAB = 327,
+     T_TTL = 328,
+     T_RRCLASS = 329,
+     URR = 330,
+     T_UTYPE = 331
    };
 #endif
 /* Tokens.  */
@@ -168,33 +169,34 @@
 #define T_MAILB 308
 #define T_MAILA 309
 #define T_DS 310
-#define T_SSHFP 311
-#define T_RRSIG 312
-#define T_NSEC 313
-#define T_DNSKEY 314
-#define T_SPF 315
-#define T_NSEC3 316
-#define T_IPSECKEY 317
-#define T_DHCID 318
-#define T_NSEC3PARAM 319
-#define DOLLAR_TTL 320
-#define DOLLAR_ORIGIN 321
-#define NL 322
-#define SP 323
-#define STR 324
-#define PREV 325
-#define BITLAB 326
-#define T_TTL 327
-#define T_RRCLASS 328
-#define URR 329
-#define T_UTYPE 330
+#define T_DLV 311
+#define T_SSHFP 312
+#define T_RRSIG 313
+#define T_NSEC 314
+#define T_DNSKEY 315
+#define T_SPF 316
+#define T_NSEC3 317
+#define T_IPSECKEY 318
+#define T_DHCID 319
+#define T_NSEC3PARAM 320
+#define DOLLAR_TTL 321
+#define DOLLAR_ORIGIN 322
+#define NL 323
+#define SP 324
+#define STR 325
+#define PREV 326
+#define BITLAB 327
+#define T_TTL 328
+#define T_RRCLASS 329
+#define URR 330
+#define T_UTYPE 331
 
 
 
 
 #if ! defined YYSTYPE && ! defined YYSTYPE_IS_DECLARED
 typedef union YYSTYPE
-#line 49 "zparser.y"
+#line 50 "zparser.y"
 {
 	domain_type	 *domain;
 	const dname_type *dname;
@@ -205,7 +207,7 @@ typedef union YYSTYPE
 	uint16_t	 *unknown;
 }
 /* Line 1489 of yacc.c.  */
-#line 209 "zparser.h"
+#line 211 "zparser.h"
 	YYSTYPE;
 # define yystype YYSTYPE /* obsolescent; will be withdrawn */
 # define YYSTYPE_IS_DECLARED 1
